import logging
from typing import List, Optional
import sys

from minionize import minionizer, ProcessCallback
from minionize.reporter import create_reporter

import pandas as pd
import time

logging.basicConfig(level=logging.INFO)


def run():
    class _Callback(ProcessCallback):
        """Simplest callback

        Takes param as a string and append them to the original command.
        """

        def to_cmd(self, param):
            return " ".join(sys.argv[1:]) + " " + str(param)

    callback = _Callback()
    m = minionizer(callback)
    m.run()


# flake8: noqa: W605
def banner():
    return """
  __  __ _       _                   _        _
 |  \/  (_)     (_)                 | |      | |
 | \  / |_ _ __  _  ___  _ __    ___| |_ __ _| |_ ___
 | |\/| | | '_ \| |/ _ \| '_ \  / __| __/ _` | __/ __|
 | |  | | | | | | | (_) | | | | \__ \ || (_| | |_\__ \\
 |_|  |_|_|_| |_|_|\___/|_| |_| |___/\__\__,_|\__|___/

"""


SPACE = " "


class Section:
    def __init__(
        self, content: str, title: Optional[str] = None, caption: str = "", sep="-"
    ):
        self.title = title
        self.content = content
        self.sep = sep
        self.caption = caption

    def __str__(self) -> str:
        s = ""
        if self.title is not None:
            s += self.title
            s += "\n"
            s += self.sep * len(self.title)
            s += "\n"
        s += self.content
        s += "\n"
        s += self.caption
        return s


def _pad(left: str, right: str, margin: str) -> str:
    """This create two columns left content padded."""
    left_lines = left.splitlines()
    padding = max([len(l) for l in left_lines] + [0])
    right_lines = right.splitlines()
    max_lines = max(len(left_lines), len(right_lines))
    left_lines.extend([""] * (max_lines - len(left_lines)))
    right_lines.extend([""] * (max_lines - len(right_lines)))
    result = []
    for l, r in zip(left_lines, right_lines):
        result.append(l.ljust(padding, SPACE) + margin + r)
    return "\n".join(result)


class Row:
    def __init__(
        self, sections: List[Section], title: Optional[str] = None, margin: int = 8
    ):
        self.title = title
        self.margin = margin
        self._margin = SPACE * margin
        self.sections = sections
        self.title = title

    def __str__(self) -> str:
        s = ""
        if self.title is not None:
            s += self.title
            s += "\n"
            s += "=" * len(self.title)
            s += "\n"

        if self.sections == []:
            return s
        str_sections = str(self.sections[0])
        for section in self.sections[1:]:
            str_sections = _pad(str_sections, str(section), self._margin)
        s += str_sections + "\n\n"
        return s


def _ascii_chart(df: pd.DataFrame, COLS: int):
    # we want two columns with some margins
    # but this shouldn't exceed COLS
    from asciichartpy import plot, blue, green, lightmagenta, colored

    margin = 3
    columns = (COLS / 2) - margin
    line = []
    derivative1 = [0]
    derivative10 = [0] * 10
    start = df.start.min()
    end = df.end.max()
    step = max(0.1, round((end - start) / columns, 1))
    current = df.start.min()
    while current < df.end.max() + step:
        # how many finished before this point
        gen_before = len(df[df.end < current])
        line.append(gen_before)
        if len(line) > 1:
            derivative1.append((line[-1] - line[-2]) / step)
        if len(line) > 10:
            derivative10.append((line[-1] - line[-11]) / (10 * step))
        current += step
    # mean rate
    mean = [len(df) / (df.end.max() - df.start.min())] * len(derivative1)

    left = Section(
        plot(line, {"height": 10}), title=f"Number of generation [{step}s step)]"
    )

    config = dict(height=10, colors=[lightmagenta, blue, green])
    right = Section(
        plot([mean, derivative1, derivative10], config),
        title=f"Generation rate (/s) ",
        caption=f"{colored('mean rate', lightmagenta)} \nsliding windows of {colored('1*step', blue)}, {colored('10*step', green)}",
    )
    print(Row([left, right]))


def _stats(df: pd.DataFrame):
    # extract process time
    def extract_process(atomic_actions):
        return [
            a["end"] - a["start"] for a in atomic_actions if a["name"] == "process"
        ][0]

    from datetime import datetime

    format = "%Y-%m-%d %H:%M:%S"
    start = datetime.fromtimestamp(df.start.min())
    end = datetime.fromtimestamp(df.end.max())
    df["process"] = df.atomic_actions.apply(extract_process)

    title = f"Processing time (first={start.strftime(format)} last={end.strftime(format)} duration={end-start})"
    s1 = Section(df.process.describe().to_string(), title="Process time summary")
    s2 = Section(
        df.groupby(pd.cut(df.process, bins=10, precision=1))
        .count()
        .process.to_string(),
        title="Process time distribution",
    )
    by_error = df.groupby("error").count().process
    content = by_error.to_string()
    if by_error.empty:
        content = "No processing error detected."

    s3 = Section(content, title="Processing errors")
    print(Row([s1, s2, s3], title=title))


def _resources(df):
    nodes = df.nodename.unique()
    extra_title = ""
    if "oar_job_id" in df:
        oar_job_ids = df.oar_job_id.unique()
        extra_title = f" / {len(oar_job_ids)} jobs"

    by = []
    if "oar_job_id" in df:
        by.append("oar_job_id")
    by.append("nodename")
    # since last mean process time
    duration = round(df.process.mean(), 0)
    last = (
        df[df.end > (time.time() - duration)].groupby(by=by).genation.count().nlargest()
    )

    # print some columns
    left = Section(
        df.groupby(by=by).genation.count().nlargest().to_string(),
        title="Most active (all time)",
    )
    if last.empty:
        content = "No data."
    else:
        content = last.to_string()

    right = Section(content, title=f"Most active (last {duration}s)")

    print(Row([left, right], title=f"Resources ({len(nodes)} nodes {extra_title})"))


def status():
    """Read the reporter backend and get some stats out of it."""
    COLS = 100
    reporter = create_reporter()
    df = reporter.to_pandas()
    if df.empty:
        print(f"{reporter} reports no stat")
        return
    nb_generation = len(df)
    rate_last = nb_generation / (df.end.max() - df.start.min())
    print(banner())
    _stats(df)
    _ascii_chart(df, COLS)
    _resources(df)



import os

# engine control
ENV_ENGINE = "MINION_ENGINE"
ENV_DRY_MODE = "DRY"
ENGINE_GOOGLE = "google"
ENGINE_EXECO = "execo"
ENGINE_PULSAR = "pulsar"

# default to some engine
ENGINE = os.getenv(ENV_ENGINE, ENGINE_EXECO)

# reporter control
ENV_REPORTER = "MINION_REPORTER"
REPORTER_STDOUT = "stdout"
REPORTER_NULL = "null"
REPORTER_JSON = "json"

# default to some reporter
REPORTER = os.getenv(ENV_REPORTER, REPORTER_NULL)

# json reporter
ENV_REPORTER_JSON_DIRECTORY = "REPORTER_JSON_DIRECTORY"
DEFAULT_JSON_REPORTER_DIRECTORY = "minion-report"



from minionize.reporter import NullReporter
import unittest
import unittest.mock as mock


from minionize import EntryPoint, ProcessCallback


class TestException(Exception):
    pass


class TestEntryPointGeneral(unittest.TestCase):
    def setUp(self):
        engine = mock.Mock()
        engine.next = mock.Mock(side_effect=["test", None])
        engine.ack = mock.Mock()
        engine.nack = mock.Mock()
        callback = mock.Mock()
        callback.setup = mock.Mock()
        callback.process = mock.Mock()
        callback.teardown = mock.Mock()

        self.engine = engine
        self.callback = callback
        self.reporter = NullReporter()

    def test_normal_operations(self):
        entrypoint = EntryPoint(self.callback, self.engine, self.reporter)
        entrypoint.run()

        self.assertEqual(2, self.engine.next.call_count)
        self.assertEqual(1, self.callback.setup.call_count)
        self.assertEqual(1, self.callback.process.call_count)
        self.assertEqual(1, self.callback.teardown.call_count)

    def test_operations_one_exception(self):
        exception = TestException("test_exception")
        self.callback.process = mock.Mock(side_effect=[exception, None])
        entrypoint = EntryPoint(self.callback, self.engine, self.reporter)
        entrypoint.run()

        self.assertEqual(2, self.engine.next.call_count)
        self.assertEqual(1, self.callback.setup.call_count)
        self.assertEqual(1, self.callback.process.call_count)
        self.assertEqual(1, self.callback.teardown.call_count)
        # check parameters
        self.callback.teardown.assert_called_with("test", self.engine, exception)


class TestEntryPointWithProcessCallback(unittest.TestCase):
    def setUp(self):
        engine = mock.Mock()
        engine.next = mock.Mock(side_effect=["test", None])
        engine.ack = mock.Mock()
        engine.nack = mock.Mock()
        callback = ProcessCallback()
        self.engine = engine
        self.callback = callback
        self.reporter = NullReporter()

    def test_normal_operations(self):
        entrypoint = EntryPoint(self.callback, self.engine, self.reporter)
        entrypoint.run()

        self.assertEqual(2, self.engine.next.call_count)
        self.assertEqual(1, self.engine.ack.call_count)
        self.assertEqual(0, self.engine.nack.call_count)
        self.engine.ack.assert_called_with("test")

    def test_operations_one_exception(self):
        class CustomCallback(ProcessCallback):
            def to_cmd(self, param):
                return "exit 1"

        entrypoint = EntryPoint(CustomCallback(), self.engine, self.reporter)
        entrypoint.run()

        self.assertEqual(2, self.engine.next.call_count)
        self.assertEqual(0, self.engine.ack.call_count)
        self.assertEqual(1, self.engine.nack.call_count)



from minionize.cli import Row

import unittest


class TestRow(unittest.TestCase):
    def test_add_row(self):
        left = Row("abc", "")
        right = Row("def", "")
        r = left + right
        self.assertEqual("abc   def", r.title)
        self.assertEqual("", r.content)

    def test_add_row_content(self):
        left_content = """a
aa
aaa
aaaa
"""
        right_content = """bbbb
bbb
bb
b
"""
        left = Row("aaaaa", left_content)
        right = Row("bbbbb", right_content)
        r = left + right
        expected_content = """a       bbbb
aa      bbb
aaa     bb
aaaa    b"""
        self.assertEqual(expected_content, r.content)



from abc import ABC, abstractclassmethod, abstractmethod
from functools import wraps
import logging
import os
import pandas as pd
from pathlib import Path
from typing import Mapping, List, Optional
from uuid import uuid4
import time

from ..env import (
    REPORTER,
    REPORTER_STDOUT,
    REPORTER_JSON,
    ENV_REPORTER_JSON_DIRECTORY,
    DEFAULT_JSON_REPORTER_DIRECTORY,
)
from ..types import Param

LOGGER = logging.getLogger(__name__)

ENV_VARS = ["OAR_JOB_ID"]


class Event:
    def __init__(self, generation: int):
        self.generation = generation
        self._start = time.time()
        self._atomic_actions: List = []
        self._end: Optional[float] = None
        self._param: Optional[Param] = None
        self._error: Optional[Exception] = None
        # capture some of the env here OAR_JOB_ID if any, hostname ...
        self._extra: Mapping = {}
        self._fill_with_env()

    def __repr__(self):
        return (
            f"<Event ({id(self)}) "
            f"generation={self.generation} "
            f"start={self._start} "
            f"end={self.end } "
            f"error={str(self.error)} "
            f"param={self.param} "
            f"atomic_actions={self._atomic_actions} >"
        )

    def _fill_with_env(self):
        for env_var in ENV_VARS:
            if os.getenv(env_var) is not None:
                self._extra.update(oar_job_id=os.getenv(env_var))
        self._extra.update(nodename=os.uname().nodename)

    @property
    def param(self) -> Param:
        return self._param

    @param.setter
    def param(self, param: Param):
        self._param = param

    @property
    def error(self) -> Optional[Exception]:
        return self._error

    @error.setter
    def error(self, error: Optional[Exception]):
        self._error = error

    @property
    def end(self) -> Optional[float]:
        return self._end

    @end.setter
    def end(self, end: float):
        self._end = end

    def to_dict(self) -> Mapping:
        e = None
        if self._error is not None:
            e = str(self._error)
        d = dict(
            generation=self.generation,
            start=self._start,
            end=self._end,
            param=self._param,
            error=e,
            atomic_actions=self._atomic_actions,
            **self._extra,
        )
        return d

    def timeit(self, f):
        """A decorator that add custom atomic action to the event.

        The rationale is to pass that to the user code, so that the user can
        register custom action.
        """

        @wraps(f)
        def wrapped(*args, **kwargs):
            start = time.time()
            r = None
            exception = None
            try:
                r = f(*args, **kwargs)
            except Exception as e:
                exception = e
            finally:
                end = time.time()
                name = getattr(f, "__name__", "unknown")
                atomic_action = dict(name=name, start=start, end=end)
                self._atomic_actions.append(atomic_action)
                if exception is None:
                    return r
                else:
                    raise exception

        return wrapped


class Reporter(ABC):
    """Base class for our reporters."""

    def __init__(self, **kwargs):
        self.extra = kwargs

    @abstractmethod
    def fire(self, event: Event):
        """Fire an event to the reporter.

        Args:
            name: name of the event
            value: content of the event
        """
        pass

    @abstractmethod
    def to_pandas(self) -> pd.DataFrame:
        return pd.DataFrame()

    @abstractclassmethod
    def from_env(cls):
        pass


class StdoutReporter(Reporter):
    def fire(self, event: Event):
        print(event)

    def to_pandas(self) -> pd.DataFrame:
        return pd.DataFrame()

    @classmethod
    def from_env(cls):
        return cls()


class NullReporter(Reporter):
    def fire(self, event: Event):
        pass

    def to_pandas(self) -> pd.DataFrame:
        return pd.DataFrame()

    @classmethod
    def from_env(cls):
        return cls()


class JsonReporter(Reporter):
    """Report events in a jsonp file."""

    def __init__(self, directory: Path, **kwargs):
        super().__init__(**kwargs)
        self.directory = directory
        self._uid = uuid4()
        self.target_file = directory / str(self._uid)
        self._buffer: List = []
        self.directory.mkdir(parents=True, exist_ok=True)

    def fire(self, event: Event):
        # force new ref
        self._buffer.append(event)
        self._flush()

    def _flush(self):
        import json

        with self.target_file.open("a+") as f:
            for event in self._buffer:
                json.dump(event.to_dict(), f)
                f.write("\n")
        self._buffer = []

    def to_pandas(self) -> pd.DataFrame:
        df = pd.DataFrame()
        for p in self.directory.glob("*"):
            LOGGER.debug(f"Reading {p}")
            df = pd.concat([df, pd.read_json(p, lines=True)])
        return df

    @classmethod
    def from_env(cls):
        import os

        directory = Path(
            os.getenv(ENV_REPORTER_JSON_DIRECTORY, DEFAULT_JSON_REPORTER_DIRECTORY)
        )
        return cls(directory)


def create_reporter() -> Reporter:
    if REPORTER == REPORTER_STDOUT:
        return StdoutReporter.from_env()

    if REPORTER == REPORTER_JSON:
        return JsonReporter.from_env()

    return NullReporter.from_env()
